'use strict';

/* eslint-env node, mocha */

const expect = require('chai').expect;
const rowResult = require('../../../lib/DevAPI/RowResult');
const Column = require('../../../lib/DevAPI/Column');

describe('RowResult', () => {
    context('fetchAll()', () => {
        it('returns an empty array when there are no items in the result-set', () => {
            expect(rowResult().fetchAll()).to.deep.equal([]);
            expect(rowResult({ results: undefined }).fetchAll()).to.deep.equal([]);
            expect(rowResult({ results: [] }).fetchAll()).to.deep.equal([]);
            expect(rowResult({ results: [[]] }).fetchAll()).to.deep.equal([]);
            expect(rowResult({ results: [null] }).fetchAll()).to.deep.equal([]);
            expect(rowResult({ results: [[null]] }).fetchAll()).to.deep.equal([]);
        });

        it('returns an array containing the data counterpart of each item in the result-set', () => {
            const expected = ['foo', 'bar'];

            expect(rowResult({ results: [[expected[0], expected[1]]] }).fetchAll()).to.deep.equal(expected);
        });

        it('is aware that fetchOne() might have been used before', () => {
            const expected = ['foo', 'bar'];
            const results = [[null, null, expected[0], expected[1]]];

            expect(rowResult({ results }).fetchAll()).to.deep.equal(expected);
        });
    });

    context('fetchOne()', () => {
        it('returns undefined when there are no items in the result-set', () => {
            /* eslint-disable no-unused-expressions */
            expect(rowResult().fetchOne()).to.not.exist;
            expect(rowResult({ results: undefined }).fetchOne()).to.not.exist;
            expect(rowResult({ results: [] }).fetchOne()).to.not.exist;
            expect(rowResult({ results: [[]] }).fetchOne()).to.not.exist;
            expect(rowResult({ results: [null] }).fetchOne()).to.not.exist;
            /* eslint-enable no-unused-expressions */
            return expect(rowResult({ results: [[null]] }).fetchOne()).to.not.exist;
        });

        it('returns the next available item in the result-set', () => {
            expect(rowResult({ results: [['foo']] }).fetchOne()).to.equal('foo');
        });

        it('deallocates the memory when a result-set item has been consumed', () => {
            const res = rowResult({ results: [[{ data: 'foo' }, { data: 'bar' }]] });

            res.fetchOne();
            expect(res.getResults()).to.deep.equal([[null, { data: 'bar' }]]);
        });

        it('deallocates the memory when the entire result-set has been consumed', () => {
            const res = rowResult({ results: [[{ data: 'foo' }]] });

            res.fetchOne();
            expect(res.getResults()).to.deep.equal([null]);
        });
    });

    context('getAffectedItemsCount()', () => {
        it('returns the same result as getAffectedRowsCount()', () => {
            expect(rowResult({ rowsAffected: 3 }).getAffectedItemsCount()).to.equal(3);
        });
    });

    context('getColumns()', () => {
        it('returns the metadata for each item in the result-set wrapped as a Column instance', () => {
            const res = rowResult({ metadata: [[{ name: 'foo' }, { name: 'bar' }]] });
            const columns = res.getColumns();

            expect(columns).to.have.lengthOf(2);
            columns.forEach(column => expect(column).to.be.an.instanceOf(Column));
        });

        it('returns an empty list if there is no metadata available in the result-set', () => {
            expect(rowResult({ results: undefined }).getColumns()).to.deep.equal([]);
            expect(rowResult({ results: [] }).getColumns()).to.deep.equal([]);
        });
    });

    context('getWarnings()', () => {
        it('returns the list of warnings generated by the server for a given operation', () => {
            const warnings = ['foo', 'bar'];

            expect(rowResult({ warnings }).getWarnings()).to.deep.equal(warnings);
        });
    });

    context('getWarningsCount()', () => {
        it('returns the number of warnings generated by the server for a given operation', () => {
            const warnings = ['foo', 'bar', 'baz'];

            expect(rowResult({ warnings }).getWarningsCount()).to.deep.equal(3);
        });
    });

    context('nextResult()', () => {
        it('returns false if there are no other result-sets available', () => {
            /* eslint-disable no-unused-expressions */
            expect(rowResult().nextResult()).to.be.false;
            expect(rowResult({ results: undefined }).nextResult()).to.be.false;
            /* eslint-enable no-unused-expressions */
            return expect(rowResult({ results: [] }).nextResult()).to.be.false;
        });

        it('moves the cursor to the next available result-set', () => {
            const res = rowResult({ results: [['foo'], ['bar']] });

            // eslint-disable-next-line no-unused-expressions
            expect(res.nextResult()).to.be.true;
            expect(res.fetchOne()).to.equal('bar');
        });
    });
});
